@@ -1,6 +1,6 @@
-'''
+"""
 Bionic-specific exception classes.
-'''
+"""
 
 
 class UndefinedEntityError(KeyError):
@@ -25,3 +25,15 @@ class UnsupportedSerializedValueError(Exception):
 
 class CodeVersioningError(Exception):
     pass
+
+
+class EntitySerializationError(Exception):
+    pass
+
+
+class EntityComputationError(Exception):
+    pass
+
+
+class AttributeValidationError(Exception):
+    pass
