@@ -1,7 +1,7 @@
-'''
+"""
 Contains the FlowBuilder and Flow classes, which implement the core workflow
 construction and execution APIs (respectively).
-'''
+"""
 
 import os
 import shutil
@@ -9,6 +9,7 @@ import warnings
 from pathlib import Path, PosixPath
 from importlib import reload
 from textwrap import dedent
+from uuid import uuid4
 
 import pyrsistent as pyrs
 import pandas as pd
@@ -18,18 +19,31 @@ from . import protocols as protos
 from .cache import LocalStore, GcsCloudStore, PersistentCache
 from .datatypes import CaseKey, VersioningPolicy
 from .exception import (
-    UndefinedEntityError, AlreadyDefinedEntityError, IncompatibleEntityError)
+    UndefinedEntityError,
+    AlreadyDefinedEntityError,
+    IncompatibleEntityError,
+)
 from .provider import (
-    ValueProvider, multi_index_from_case_keys, as_provider,
-    provider_wrapper, AttrUpdateProvider)
+    ValueProvider,
+    multi_index_from_case_keys,
+    as_provider,
+    provider_wrapper,
+    AttrUpdateProvider,
+)
 from .deriver import EntityDeriver
+from .descriptors import DescriptorNode
 from . import decorators
 from .util import (
-    group_pairs, check_exactly_one_present, check_at_most_one_present,
-    copy_to_gcs, FileCopier, oneline
+    group_pairs,
+    check_exactly_one_present,
+    check_at_most_one_present,
+    copy_to_gcs,
+    FileCopier,
+    oneline,
 )
 
 import logging
+
 logger = logging.getLogger(__name__)
 
 DEFAULT_PROTOCOL = protos.CombinedProtocol(
@@ -148,9 +162,13 @@ class FlowState(pyrs.PClass):
             joint_names = provider.get_joint_names()
             for related_name in joint_names:
                 if related_name not in names:
-                    raise IncompatibleEntityError(oneline(f'''
+                    raise IncompatibleEntityError(
+                        oneline(
+                            f"""
                         Can't remove cases for entity {name!r} without also
-                        removing entities {joint_names!r}'''))
+                        removing entities {joint_names!r}"""
+                        )
+                    )
 
             # Delete it.
             state = state._erase_provider(name)
@@ -168,11 +186,11 @@ class FlowState(pyrs.PClass):
     def _erase_provider(self, name):
         state = self
         if name in state.providers_by_name:
-            state = state.set(
-                providers_by_name=state.providers_by_name.remove(name))
+            state = state.set(providers_by_name=state.providers_by_name.remove(name))
         if name in state.default_provider_names:
-            state = state.set(default_provider_names=(
-                state.default_provider_names.remove(name)))
+            state = state.set(
+                default_provider_names=(state.default_provider_names.remove(name))
+            )
         return state
 
     def _set_provider(self, provider):
@@ -180,7 +198,8 @@ class FlowState(pyrs.PClass):
         for name in provider.attrs.names:
             state = state._erase_provider(name)
             state = state.set(
-                providers_by_name=state.providers_by_name.set(name, provider))
+                providers_by_name=state.providers_by_name.set(name, provider)
+            )
         return state
 
 
@@ -256,17 +275,15 @@ class FlowBuilder(object):
             check_at_most_one_present(doc=doc, docstring=docstring)
             warnings.warn(
                 "The `docstring` argument to `FlowBuilder.declare` is "
-                "deprecated; use `doc` instead.")
+                "deprecated; use `doc` instead."
+            )
             doc = docstring
 
         self._state = self._state.create_provider(name, protocol, doc)
 
-    def assign(self, name,
-               value=None,
-               values=None,
-               protocol=None,
-               doc=None,
-               docstring=None):
+    def assign(
+        self, name, value=None, values=None, protocol=None, doc=None, docstring=None
+    ):
         """
         Creates a new entity and assigns it a value.
 
@@ -300,7 +317,8 @@ class FlowBuilder(object):
             check_at_most_one_present(doc=doc, docstring=docstring)
             warnings.warn(
                 "The `docstring` argument to `FlowBuilder.assign` is "
-                "deprecated; use `doc` instead.")
+                "deprecated; use `doc` instead."
+            )
             doc = docstring
 
         for value in values:
@@ -344,7 +362,7 @@ class FlowBuilder(object):
         provider = state.get_provider(name)
         # This provider must have a single name and single protocol; otherwise
         # we wouldn't have been able to clear it.
-        protocol, = provider.attrs.protocols
+        (protocol,) = provider.attrs.protocols
 
         protocol.validate(value)
 
@@ -433,10 +451,14 @@ class FlowBuilder(object):
         for name, value in name_value_pairs:
             provider = state.get_provider(name)
             if len(provider.attrs.protocols) > 1:
-                raise IncompatibleEntityError(oneline(f'''
+                raise IncompatibleEntityError(
+                    oneline(
+                        f"""
                     Can't add case for entity co-generated with other
-                    entities {provider.attrs.names!r}'''))
-            protocol, = provider.attrs.protocols
+                    entities {provider.attrs.names!r}"""
+                    )
+                )
+            (protocol,) = provider.attrs.protocols
             protocol.validate(value)
             token = protocol.tokenize(value)
 
@@ -487,7 +509,7 @@ class FlowBuilder(object):
 
         self._state = self._state.delete_providers(names)
 
-    def merge(self, flow, keep='error', allow_name_match=False):
+    def merge(self, flow, keep="error", allow_name_match=False):
         """
         Updates this builder by importing all entities from another flow.
 
@@ -534,101 +556,110 @@ class FlowBuilder(object):
         # guess it's not a huge deal.  But overall, the way we store a flow's
         # name might deserve to be revisited later.
         if not allow_name_match:
-            old_name_provider = old_state.providers_by_name.get(
-                'core__flow_name')
+            old_name_provider = old_state.providers_by_name.get("core__flow_name")
             if (
-                    old_name_provider is not None and
-                    isinstance(old_name_provider, ValueProvider) and
-                    len(old_name_provider._values_by_case_key.keys()) == 1):
-                old_flow_name, = old_name_provider._values_by_case_key.values()
+                old_name_provider is not None
+                and isinstance(old_name_provider, ValueProvider)
+                and len(old_name_provider._values_by_case_key.keys()) == 1
+            ):
+                (old_flow_name,) = old_name_provider._values_by_case_key.values()
                 new_flow_name = flow.name
                 if old_flow_name == new_flow_name and not allow_name_match:
-                    raise ValueError(oneline(f'''
+                    raise ValueError(
+                        oneline(
+                            f"""
                         Attempting to merge two flows with the same name
                         ({new_flow_name!r}).
                         Sharing names between flows is generally unwise,
                         since they will then also share the same cache space;
                         however, you can disable this check by passing
-                        ``allow_name_match=True``.'''))
+                        ``allow_name_match=True``."""
+                        )
+                    )
 
         # Identify all the names that could appear in the merged flow, and
         # associate each one with a potential conflict.
         all_names = set(old_state.providers_by_name.keys()).union(
-                new_state.providers_by_name.keys())
+            new_state.providers_by_name.keys()
+        )
         conflicts_by_name = {
-            name: MergeConflict(
-                old_state=old_state, new_state=new_state, name=name)
+            name: MergeConflict(old_state=old_state, new_state=new_state, name=name)
             for name in all_names
         }
 
         # Resolve each conflict individually.
         for conflict in conflicts_by_name.values():
             if conflict.old_provider is None:
-                conflict.resolve('new', 'no conflicting definition')
+                conflict.resolve("new", "no conflicting definition")
                 continue
 
             if conflict.new_provider is None:
-                conflict.resolve('old', 'no conflicting definition')
+                conflict.resolve("old", "no conflicting definition")
                 continue
 
-            if conflict.name == 'core__flow_name':
-                conflict.resolve('old', 'flow name is never merged')
+            if conflict.name == "core__flow_name":
+                conflict.resolve("old", "flow name is never merged")
                 continue
 
             if conflict.new_is_default:
-                conflict.resolve('old', 'conflicting definition is default')
+                conflict.resolve("old", "conflicting definition is default")
                 continue
 
             if conflict.old_is_default:
-                conflict.resolve('new', 'conflicting definition is default')
+                conflict.resolve("new", "conflicting definition is default")
                 continue
 
             if conflict.old_protocol is conflict.new_protocol:
                 if conflict.new_is_only_declaration:
                     conflict.resolve(
-                        'old',
-                        'conflicting definition has matching protocol and '
-                        'no value')
+                        "old",
+                        "conflicting definition has matching protocol and " "no value",
+                    )
                     continue
                 elif conflict.old_is_only_declaration:
                     conflict.resolve(
-                        'new',
-                        'conflicting definition has matching protocol and '
-                        'no value')
+                        "new",
+                        "conflicting definition has matching protocol and " "no value",
+                    )
                     continue
 
-            if keep == 'error':
-                raise AlreadyDefinedEntityError(oneline(f'''
+            if keep == "error":
+                raise AlreadyDefinedEntityError(
+                    oneline(
+                        f"""
                     Merge failure: Entity {conflict.name!r} exists in both
                     old and new flows;
-                    use the ``keep`` argument to specify which to keep'''))
+                    use the ``keep`` argument to specify which to keep"""
+                    )
+                )
 
-            elif keep == 'self':
-                conflict.resolve('old', 'keep=self')
+            elif keep == "self":
+                conflict.resolve("old", "keep=self")
                 continue
 
-            elif keep == 'arg':
-                conflict.resolve('new', 'keep=arg')
+            elif keep == "arg":
+                conflict.resolve("new", "keep=arg")
                 continue
 
-            elif keep == 'old':
-                conflict.resolve('old', 'keep=old')
+            elif keep == "old":
+                conflict.resolve("old", "keep=old")
                 continue
 
-            elif keep == 'new':
-                conflict.resolve('new', 'keep=new')
+            elif keep == "new":
+                conflict.resolve("new", "keep=new")
                 continue
 
             raise ValueError(
                 "Value of ``keep`` must be one of {'error', 'self', 'arg'}; "
-                f"got {keep!r}")
+                f"got {keep!r}"
+            )
 
         # For both states, check that each jointly-defined name group is kept
         # or discarded as a whole.
         for state_name, state in [
-                    ('old', old_state),
-                    ('new', new_state),
-                ]:
+            ("old", old_state),
+            ("new", new_state),
+        ]:
             for provider in state.providers_by_name.values():
                 names = provider.get_joint_names()
                 if len(names) == 1:
@@ -636,42 +667,44 @@ class FlowBuilder(object):
 
                 conflicts = [conflicts_by_name[name] for name in names]
                 kept_conflicts = [
-                    conflict for conflict in conflicts
+                    conflict
+                    for conflict in conflicts
                     if conflict.resolution == state_name
                 ]
                 discarded_conflicts = [
-                    conflict for conflict in conflicts
+                    conflict
+                    for conflict in conflicts
                     if conflict.resolution != state_name
                 ]
                 if kept_conflicts and discarded_conflicts:
-                    kept = ', '.join(
-                        f'{c.name} ({c.reason})'
-                        for c in kept_conflicts)
-                    discarded = ', '.join(
-                        f'{c.name} ({c.reason})'
-                        for c in discarded_conflicts)
-                    raise IncompatibleEntityError(oneline(f'''
+                    kept = ", ".join(f"{c.name} ({c.reason})" for c in kept_conflicts)
+                    discarded = ", ".join(
+                        f"{c.name} ({c.reason})" for c in discarded_conflicts
+                    )
+                    raise IncompatibleEntityError(
+                        oneline(
+                            f"""
                         Merge failure: Names {names!r} in {state_name} state
                         are defined jointly and must be kept or discarded
                         together,
                         but merge logic dictates that we keep [{kept}] and
                         discard [{discarded}];
                         you should manually remove some of these names from
-                        one of the flows before merging'''))
+                        one of the flows before merging"""
+                        )
+                    )
 
         # Now we start building up our final, merged state.
         cur_state = old_state
 
         conflicts_keeping_new = [
-            conflict for conflict in conflicts_by_name.values()
-            if conflict.resolution == 'new'
+            conflict
+            for conflict in conflicts_by_name.values()
+            if conflict.resolution == "new"
         ]
 
         # First, delete all old providers that collide with our incoming ones.
-        names_to_delete = [
-            conflict.name
-            for conflict in conflicts_keeping_new
-        ]
+        names_to_delete = [conflict.name for conflict in conflicts_keeping_new]
         try:
             cur_state = cur_state.delete_providers(names_to_delete)
         except IncompatibleEntityError as e:
@@ -680,20 +713,20 @@ class FlowBuilder(object):
         # Then install each new provider -- keeping in mind that a provider
         # may have multiple names and hence multiple conflicts, but should be
         # installed exactly once.
-        providers_to_install = (
-            {
-                tuple(conflict.new_provider.attrs.names): conflict.new_provider
-                for conflict in conflicts_keeping_new
-            }.values())
+        providers_to_install = {
+            tuple(conflict.new_provider.attrs.names): conflict.new_provider
+            for conflict in conflicts_keeping_new
+        }.values()
         for provider in providers_to_install:
             # For function providers, we attach the name of their original flow
             # so that their cached data won't be confused with whatever we
             # had before.
             if (
-                    not isinstance(provider, ValueProvider) and
-                    provider.attrs.orig_flow_name is None):
+                not isinstance(provider, ValueProvider)
+                and provider.attrs.orig_flow_name is None
+            ):
                 provider = provider_wrapper(
-                    AttrUpdateProvider, 'orig_flow_name', new_flow_name
+                    AttrUpdateProvider, "orig_flow_name", new_flow_name
                 )(provider)
             cur_state = cur_state.install_provider(provider)
 
@@ -725,43 +758,56 @@ class FlowBuilder(object):
         if provider.attrs.docs is None:
             docs = [None] * len(provider.attrs.names)
             provider = decorators.docs(*docs)(provider)
-        if provider.attrs.should_persist is None:
+        if provider.attrs._can_persist is None:
             provider = decorators.persist(True)(provider)
-        if provider.attrs.should_memoize is None:
+        if provider.attrs._can_memoize is None:
             provider = decorators.memoize(True)(provider)
+        if provider.attrs.changes_per_run is None:
+            provider = decorators.changes_per_run(False)(provider)
 
-        if not (
-                provider.attrs.should_persist or
-                provider.attrs.should_memoize):
-            raise ValueError(oneline(f'''
+        if not (provider.attrs.should_persist() or provider.attrs.should_memoize()):
+            raise ValueError(
+                oneline(
+                    f"""
                 Attempted to set both persist and memoize to False.
                 At least one form of storage must be enabled for entities:
-                {func_or_provider.attrs.names!r}'''))
+                {func_or_provider.attrs.names!r}"""
+                )
+            )
         if len(provider.attrs.protocols) != len(provider.attrs.names):
             names = provider.attrs.names
             protocols = provider.attrs.protocols
-            raise ValueError(oneline(f'''
+            raise ValueError(
+                oneline(
+                    f"""
                 Number of protocols must match the number of names;
                 got {len(names)} names {tuple(names)!r} and
-                {len(protocols)} protocols {tuple(protocols)!r}'''))
+                {len(protocols)} protocols {tuple(protocols)!r}"""
+                )
+            )
         if len(provider.attrs.docs) != len(provider.attrs.names):
             names = provider.attrs.names
             docs = provider.attrs.docs
-            message = oneline(f'''
+            message = oneline(
+                f"""
                 Number of docs must match the number of names;
                 got {len(names)} names {tuple(names)!r} and
-                {len(docs)} docs {tuple(docs)!r}''')
+                {len(docs)} docs {tuple(docs)!r}"""
+            )
             if len(provider.attrs.docs) == 1:
-                prefix = oneline('''
+                prefix = oneline(
+                    """
                     Using a single doc string for a multi-output function is
                     deprecated and will become an error condition in a future
                     release; use the ``@docs`` decorator to specify
-                    one doc string for each entity.  Details:''')
-                warnings.warn(prefix + '\n' + message)
+                    one doc string for each entity.  Details:"""
+                )
+                warnings.warn(prefix + "\n" + message)
 
                 docs = [docs[0]] * len(names)
                 provider = AttrUpdateProvider(
-                    provider, 'docs', docs, allow_override=True)
+                    provider, "docs", docs, allow_override=True
+                )
             else:
                 raise ValueError(message)
 
@@ -781,7 +827,8 @@ class FlowBuilder(object):
         warnings.warn(
             "FlowBuilder.derive is deprecated and will be repurposed in the "
             "future; use FlowBuilder.__call__ (i.e., using the builder as a "
-            "decorator) instead.")
+            "decorator) instead."
+        )
 
         return self.derive(func_or_provider)
 
@@ -796,7 +843,7 @@ class FlowBuilder(object):
         return cls(name=None, _state=FlowState())
 
     def _set_name(self, name):
-        self.set('core__flow_name', name)
+        self.set("core__flow_name", name)
 
     def _set_for_case_key(self, case_key, name, value):
         self._state = self._state.add_case(name, case_key, value)
@@ -804,8 +851,7 @@ class FlowBuilder(object):
     def _set_for_last_case(self, name, value):
         last_case_key = self._state.last_added_case_key
         if last_case_key is None:
-            raise ValueError(
-                "A case must have been added before calling this method")
+            raise ValueError("A case must have been added before calling this method")
 
         self._set_for_case_key(last_case_key, name, value)
 
@@ -840,9 +886,7 @@ class MergeConflict(object):
         return self._provider_is_only_declaration(self.new_provider)
 
     def _provider_is_only_declaration(self, provider):
-        return (
-            isinstance(provider, ValueProvider) and
-            not provider.has_any_cases())
+        return isinstance(provider, ValueProvider) and not provider.has_any_cases()
 
     def resolve(self, resolution, reason):
         self.resolution = resolution
@@ -855,6 +899,7 @@ class FlowCase(object):
 
     These should be constructed by the ``FlowBuilder`` object, not by users.
     """
+
     def __init__(self, builder, key):
         self.key = key
         self._builder = builder
@@ -930,8 +975,8 @@ class Flow(object):
             The name of an entity.
         """
         warnings.warn(
-            "`Flow.entity_docstring` is deprecated; "
-            "use `Flow.entity_doc` instead.")
+            "`Flow.entity_docstring` is deprecated; " "use `Flow.entity_doc` instead."
+        )
         return self.entity_doc(name)
 
     def to_builder(self):
@@ -987,23 +1032,29 @@ class Flow(object):
         The value of the entity, or a collection containing its values.
         """
 
-        result_group = self._deriver.derive(name)
-        if mode is object or mode == 'object':
+        dnode = DescriptorNode.from_descriptor(name)
+        result_group = self._deriver.derive(dnode)
+
+        if mode is object or mode == "object":
             values = [result.value for result in result_group]
         else:
             # all other modes expect the entity to be persisted
             result_file_paths = [result.file_path for result in result_group]
 
             if None in result_file_paths:
-                raise ValueError(oneline(f'''
+                raise ValueError(
+                    oneline(
+                        f"""
                     Entity {name!r} is not persisted but persisted file is
-                    expected by mode {mode!r}'''))
+                    expected by mode {mode!r}"""
+                    )
+                )
 
-            if mode is Path or mode == 'path':
+            if mode is Path or mode == "path":
                 values = result_file_paths
-            elif mode == 'FileCopier':
+            elif mode == "FileCopier":
                 values = [FileCopier(fp) for fp in result_file_paths]
-            elif mode == 'filename':
+            elif mode == "filename":
                 values = [str(fp) for fp in result_file_paths]
             else:
                 raise ValueError(f"Unrecognized mode {mode!r}")
@@ -1013,7 +1064,8 @@ class Flow(object):
         if fmt:
             warnings.warn(
                 "The fmt argument is deprecated and will be removed in a future release. "
-                "Please use collection as a replacement.")
+                "Please use collection as a replacement."
+            )
 
         collection = fmt or collection
 
@@ -1023,24 +1075,19 @@ class Flow(object):
             if len(values) > 1:
                 raise ValueError(f"Entity {name!r} has multiple values")
             return values[0]
-        elif collection is list or collection == 'list':
+        elif collection is list or collection == "list":
             return values
-        elif collection is set or collection == 'set':
+        elif collection is set or collection == "set":
             return set(values)
-        elif collection is pd.Series or collection == 'series':
+        elif collection is pd.Series or collection == "series":
             if len(result_group.key_space) > 0:
                 index = multi_index_from_case_keys(
-                    case_keys=[
-                        result.query.case_key for result in result_group],
+                    case_keys=[result.query.case_key for result in result_group],
                     ordered_key_names=list(result_group.key_space),
                 )
             else:
                 index = None
-            return pd.Series(
-                name=name,
-                data=values,
-                index=index,
-            )
+            return pd.Series(name=name, data=values, index=index,)
         else:
             raise ValueError(f"Unrecognized collection type {collection!r}")
 
@@ -1069,15 +1116,21 @@ class Flow(object):
         """
 
         warnings.warn(
-            "Flow#export is deprecated and the same functionality is available through Flow#get.")
+            "Flow#export is deprecated; use the mode argument to Flow#get " "instead"
+        )
 
-        result_group = self._deriver.derive(name)
+        dnode = DescriptorNode.from_descriptor(name)
+        result_group = self._deriver.derive(dnode)
         if len(result_group) != 1:
-            raise ValueError(oneline(f'''
+            raise ValueError(
+                oneline(
+                    f"""
                 Can only export an entity if it has a single value;
-                entity {name!r} has {len(result_group)} values'''))
+                entity {name!r} has {len(result_group)} values"""
+                )
+            )
 
-        result, = result_group
+        (result,) = result_group
 
         if result.file_path is None:
             raise ValueError(f"Entity {name!r} is not locally persisted")
@@ -1096,15 +1149,14 @@ class Flow(object):
             dst_file_path = Path(file_path)
             dst_dir_path = dst_file_path.parent
 
-        if not dst_dir_path.exists() and 'gs:/' not in str(dst_dir_path):
+        if not dst_dir_path.exists() and "gs:/" not in str(dst_dir_path):
             dst_dir_path.mkdir(parents=True)
 
         dst_file_path_str = str(dst_file_path)
 
-        if dst_file_path_str.startswith('gs:/'):
+        if dst_file_path_str.startswith("gs:/"):
             # The path object combines // into /, so we revert it here
-            copy_to_gcs(
-                str(src_file_path), dst_file_path_str.replace('gs:/', 'gs://'))
+            copy_to_gcs(str(src_file_path), dst_file_path_str.replace("gs:/", "gs://"))
         else:
             shutil.copyfile(str(src_file_path), dst_file_path_str)
 
@@ -1120,8 +1172,9 @@ class Flow(object):
         Like ``FlowBuilder.assign``, but returns a new copy of this flow.
         """
 
-        return self._updating(lambda builder: builder.assign(
-            name, value, values, protocol))
+        return self._updating(
+            lambda builder: builder.assign(name, value, values, protocol)
+        )
 
     def setting(self, name, value=None, values=None):
         """
@@ -1144,8 +1197,7 @@ class Flow(object):
         Use after calling ``Flow.adding_case``.
         """
 
-        return self._updating(
-            lambda builder: builder._set_for_last_case(name, value))
+        return self._updating(lambda builder: builder._set_for_last_case(name, value))
 
     def clearing_cases(self, *names):
         """
@@ -1154,7 +1206,7 @@ class Flow(object):
 
         return self._updating(lambda builder: builder.clear_cases(*names))
 
-    def merging(self, flow, keep='error'):
+    def merging(self, flow, keep="error"):
         """
         Like ``FlowBuilder.merge``, but returns a new copy of this flow.
         """
@@ -1165,11 +1217,12 @@ class Flow(object):
     def name(self):
         """Returns the name of this flow."""
 
-        return self.get('core__flow_name')
+        return self.get("core__flow_name")
 
     def render_dag(self, include_core=False, vertical=False, curvy_lines=False):
         """
-        Returns an image with a visualization of this flow's DAG.
+        Returns a ``FlowImage`` with a visualization of this flow's DAG. This
+        object behaves similarly to a Pillow ``Image`` object.
 
         Will fail if Graphviz is not installed on the system.
         """
@@ -1177,9 +1230,10 @@ class Flow(object):
         from . import dagviz
 
         graph = self._deriver.export_dag(include_core)
-        dot = dagviz.dot_from_graph(graph, vertical, curvy_lines)
-        image = dagviz.image_from_dot(dot)
-        return image
+        dot = dagviz.dot_from_graph(
+            graph=graph, vertical=vertical, curvy_lines=curvy_lines, name=self.name,
+        )
+        return dagviz.FlowImage(dot)
 
     # TODO Should we offer an in-place version of this?  It's contrary to the
     # idea of an immutable API, but it might be more natural for the user, and
@@ -1235,7 +1289,8 @@ class Flow(object):
         if not state.is_blessed:
             raise ValueError(
                 "A flow can only be reloaded if it's the first flow built "
-                "from its builder and it hasn't been modified")
+                "from its builder and it hasn't been modified"
+            )
 
         self_name = self.name
 
@@ -1264,19 +1319,31 @@ class Flow(object):
 
         if len(blessed_candidate_flows) == 0:
             if len(unblessed_candidate_flows) > 0:
-                raise Exception(oneline(f'''
+                raise Exception(
+                    oneline(
+                        f"""
                     Found a matching flow, but it had been modified:
-                    {self_name!r}'''))
+                    {self_name!r}"""
+                    )
+                )
             else:
-                raise Exception(oneline(f'''
+                raise Exception(
+                    oneline(
+                        f"""
                     Couldn't find any flow named {self_name!r}
-                    in modules {module_names!r}'''))
+                    in modules {module_names!r}"""
+                    )
+                )
         if len(blessed_candidate_flows) > 1:
-            raise Exception(oneline(f'''
+            raise Exception(
+                oneline(
+                    f"""
                 Too many flows named {self_name!r}
                 in modules {module_names!r};
-                found {len(blessed_candidate_flows)}, wanted 1'''))
-        flow, = blessed_candidate_flows
+                found {len(blessed_candidate_flows)}, wanted 1"""
+                )
+            )
+        (flow,) = blessed_candidate_flows
 
         return flow
 
@@ -1290,10 +1357,12 @@ class Flow(object):
         if not _official:
             raise ValueError(
                 "Don't construct this class directly; "
-                "use one of the classmethod constructors")
+                "use one of the classmethod constructors"
+            )
 
+        self._uuid = str(uuid4())
         self._state = state
-        self._deriver = EntityDeriver(state)
+        self._deriver = EntityDeriver(state, self._uuid)
 
         # We replace the `get` and `setting` methods with wrapper classes
         # that have an attribute for each entity.  This allows a convenient,
@@ -1310,7 +1379,9 @@ class Flow(object):
 
             def __init__(self):
                 super(GetMethod, self).__init__(
-                    orig_get_method, 'Computes the value(s) for "{name}"')
+                    orig_get_method, 'Computes the value(s) for "{name}"'
+                )
+
         self.get = GetMethod()
 
         class SettingMethod(ShortcutProxy):
@@ -1319,9 +1390,12 @@ class Flow(object):
             def __init__(self):
                 super(SettingMethod, self).__init__(
                     orig_setting_method,
-                    oneline('''
+                    oneline(
+                        '''
                     Returns a copy of this flow with an updated value of
-                    "{name}"'''))
+                    "{name}"'''
+                    ),
+                )
 
         self.setting = SettingMethod()
 
@@ -1332,7 +1406,7 @@ class Flow(object):
 
 
 class ShortcutProxy(object):
-    '''
+    """
     Wraps a method on a Flow object to allow it to be called via an alternative
     style.
 
@@ -1348,7 +1422,7 @@ class ShortcutProxy(object):
 
     The advantage of the alternative style is that it can be autocompleted in
     IPython, Jupyter, etc.
-    '''
+    """
 
     def __init__(self, wrapped_method, docstring_prefix_template):
         self._wrapped_method = wrapped_method
@@ -1372,8 +1446,7 @@ class ShortcutProxy(object):
 
         # Set a useful docstring for this function.
         # First, the prefix explains what the function actually does.
-        doc_prefix =\
-            self._docstring_prefix_template.format(name=name)
+        doc_prefix = self._docstring_prefix_template.format(name=name)
 
         # If the related entity has any documentation, we include that
         # afterwards.
@@ -1389,7 +1462,8 @@ class ShortcutProxy(object):
             f"{main_docstring}"
             "\n\n"
             "This function is equivalent to "
-            f"``{method_name}('{name}', *args, **kwargs)``")
+            f"``{method_name}('{name}', *args, **kwargs)``"
+        )
 
         return partial
 
@@ -1398,56 +1472,59 @@ class ShortcutProxy(object):
 def create_default_flow_state():
     builder = FlowBuilder._with_empty_state()
 
-    builder.declare('core__flow_name')
+    builder.declare("core__flow_name")
 
-    builder.assign('core__persistent_cache__global_dir', 'bndata')
-    builder.assign('core__versioning_mode', 'manual')
+    builder.assign("core__persistent_cache__global_dir", "bndata")
+    builder.assign("core__versioning_mode", "manual")
 
     @builder
     @decorators.immediate
     def core__versioning_policy(core__versioning_mode):
-        if core__versioning_mode == 'manual':
+        if core__versioning_mode == "manual":
             return VersioningPolicy(
-                treat_bytecode_as_functional=False,
-                check_for_bytecode_errors=False,
+                treat_bytecode_as_functional=False, check_for_bytecode_errors=False,
             )
-        elif core__versioning_mode == 'assist':
+        elif core__versioning_mode == "assist":
             return VersioningPolicy(
-                treat_bytecode_as_functional=False,
-                check_for_bytecode_errors=True,
+                treat_bytecode_as_functional=False, check_for_bytecode_errors=True,
             )
-        elif core__versioning_mode == 'auto':
+        elif core__versioning_mode == "auto":
             return VersioningPolicy(
-                treat_bytecode_as_functional=True,
-                check_for_bytecode_errors=False,
+                treat_bytecode_as_functional=True, check_for_bytecode_errors=False,
             )
         else:
-            raise ValueError(oneline(f'''
+            raise ValueError(
+                oneline(
+                    f"""
                 core__versioning_mode must be one of
                 ('manual', 'assist', 'auto');
-                got {core__versioning_mode!r}'''))
+                got {core__versioning_mode!r}"""
+                )
+            )
 
     @builder
     @decorators.immediate
     def core__persistent_cache__flow_dir(
-            core__persistent_cache__global_dir, core__flow_name):
-        return os.path.join(
-            core__persistent_cache__global_dir, core__flow_name)
+        core__persistent_cache__global_dir, core__flow_name
+    ):
+        return os.path.join(core__persistent_cache__global_dir, core__flow_name)
 
-    builder.assign('core__persistent_cache__gcs__bucket_name', None)
-    builder.assign('core__persistent_cache__gcs__enabled', False)
+    builder.assign("core__persistent_cache__gcs__bucket_name", None)
+    builder.assign("core__persistent_cache__gcs__enabled", False)
 
     @builder
     @decorators.immediate
     def core__persistent_cache__gcs__object_path():
         import getpass
-        return f'{getpass.getuser()}/bndata/'
+
+        return f"{getpass.getuser()}/bndata/"
 
     @builder
     @decorators.immediate
     def core__persistent_cache__gcs__url(
-            core__persistent_cache__gcs__bucket_name,
-            core__persistent_cache__gcs__object_path):
+        core__persistent_cache__gcs__bucket_name,
+        core__persistent_cache__gcs__object_path,
+    ):
         bucket_name = core__persistent_cache__gcs__bucket_name
         object_path_str = core__persistent_cache__gcs__object_path
 
@@ -1455,29 +1532,26 @@ def create_default_flow_state():
             return None
 
         path = PosixPath(bucket_name) / object_path_str
-        return f'gs://{path}'
+        return f"gs://{path}"
 
     @builder
     @decorators.immediate
-    def core__persistent_cache__local_store(
-                core__persistent_cache__flow_dir,
-            ):
+    def core__persistent_cache__local_store(core__persistent_cache__flow_dir,):
         local_flow_dir = core__persistent_cache__flow_dir
         return LocalStore(local_flow_dir)
 
     @builder
     @decorators.immediate
     def core__persistent_cache__cloud_store(
-                core__persistent_cache__gcs__url,
-                core__persistent_cache__gcs__enabled,
-            ):
+        core__persistent_cache__gcs__url, core__persistent_cache__gcs__enabled,
+    ):
         gcs_url = core__persistent_cache__gcs__url
         gcs_enabled = core__persistent_cache__gcs__enabled
         if gcs_enabled:
             if gcs_url is None:
                 raise AssertionError(
-                    'core__persistent_cache__gcs__url is None, '
-                    'but needs a value')
+                    "core__persistent_cache__gcs__url is None, " "but needs a value"
+                )
             return GcsCloudStore(gcs_url)
         else:
             return None
@@ -1485,9 +1559,8 @@ def create_default_flow_state():
     @builder
     @decorators.immediate
     def core__persistent_cache(
-                core__persistent_cache__local_store,
-                core__persistent_cache__cloud_store,
-            ):
+        core__persistent_cache__local_store, core__persistent_cache__cloud_store,
+    ):
         return PersistentCache(
             local_store=core__persistent_cache__local_store,
             cloud_store=core__persistent_cache__cloud_store,
