@@ -1,4 +1,4 @@
-'''
+"""
 Contains bionic's concept of a "protocol": an object that specifies how to
 handle certain types of data.  A protocol is similar to a Python type, but
 might be more or less specific.  For example, PicklableProtocol can be used
@@ -7,7 +7,7 @@ takes additional arguments that constrain the structure of the DataFrames it
 handles.  Protocols specify how to serialize, deserialize, and validate data.
 
 This module contains a BaseProtocol class and various subclasses.
-'''
+"""
 
 from collections import Counter
 import pickle
@@ -23,21 +23,27 @@ from pyarrow import parquet, Table
 import pandas as pd
 
 from .exception import UnsupportedSerializedValueError
-from .provider import (
-    provider_wrapper, ProtocolUpdateProvider, is_func_or_provider)
+from .provider import provider_wrapper, ProtocolUpdateProvider, is_func_or_provider
 from .optdep import import_optional_dependency
 from .util import (
-    read_hashable_bytes_from_file_or_dir, oneline, recursively_copy_path,
-    single_element)
+    read_hashable_bytes_from_file_or_dir,
+    oneline,
+    recursively_copy_path,
+    single_element,
+)
 from . import tokenization
 
 
 def check_is_like_protocol(obj):
-    for method_name in 'validate', 'read', 'write':
+    for method_name in "validate", "read", "write":
         if not hasattr(obj, method_name):
-            raise ValueError(oneline(f'''
+            raise ValueError(
+                oneline(
+                    f"""
                 Expected {obj!r} to be a kind of Protocol, but didn't find
-                expected method {method_name!r}'''))
+                expected method {method_name!r}"""
+                )
+            )
 
 
 class BaseProtocol(object):
@@ -61,10 +67,7 @@ class BaseProtocol(object):
         return extension == self.get_fixed_file_extension()
 
     def supports_value(self, value):
-        return any(
-            isinstance(value, type_)
-            for type_ in self.get_all_supported_types()
-        )
+        return any(isinstance(value, type_) for type_ in self.get_all_supported_types())
 
     def write(self, value, path):
         raise NotImplementedError()
@@ -74,8 +77,12 @@ class BaseProtocol(object):
 
     SIMPLE_TYPES = {
         bool,
-        str, bytes, str,
-        int, int, float,
+        str,
+        bytes,
+        str,
+        int,
+        int,
+        float,
     }
 
     def tokenize(self, value):
@@ -84,8 +91,11 @@ class BaseProtocol(object):
         else:
             return tokenization.tokenize(value, self._write_to_bytes)
 
+    def tokenize_file(self, path):
+        return tokenization.tokenize(path, read_hashable_bytes_from_file_or_dir)
+
     def _write_to_bytes(self, value):
-        file_name = 'temp_file'
+        file_name = "temp_file"
         temp_dir = Path(tempfile.mkdtemp())
         try:
             temp_file_path = temp_dir / file_name
@@ -119,13 +129,18 @@ class BaseProtocol(object):
         if func_or_provider is not None:
             if len(kwargs) > 0:
                 raise ValueError(
-                    f"{self} can't be called with both a function and keywords")
+                    f"{self} can't be called with both a function and keywords"
+                )
             if not is_func_or_provider(func_or_provider):
-                raise ValueError(oneline(f'''
+                raise ValueError(
+                    oneline(
+                        f"""
                     {self} must be used either (a) directly as a decorator or
                     (b) with keyword arguments;
                     it can't take positional arguments.
-                    '''))
+                    """
+                    )
+                )
 
             wrapper = provider_wrapper(ProtocolUpdateProvider, self)
             return wrapper(func_or_provider)
@@ -133,7 +148,7 @@ class BaseProtocol(object):
             return self.__class__(**kwargs)
 
     def __repr__(self):
-        return f'{self.__class__.__name__}(...)'
+        return f"{self.__class__.__name__}(...)"
 
 
 class PicklableProtocol(BaseProtocol):
@@ -143,14 +158,14 @@ class PicklableProtocol(BaseProtocol):
     """
 
     def get_fixed_file_extension(self):
-        return 'pkl'
+        return "pkl"
 
     def write(self, value, path):
-        with path.open('wb') as file_:
+        with path.open("wb") as file_:
             pickle.dump(value, file_)
 
     def read(self, path, extension):
-        with path.open('rb') as file_:
+        with path.open("rb") as file_:
             return pickle.load(file_)
 
 
@@ -163,7 +178,7 @@ class DillableProtocol(BaseProtocol):
     """
 
     def get_fixed_file_extension(self):
-        return 'dill'
+        return "dill"
 
     def __init__(self, suppress_dill_side_effects=True):
         super(DillableProtocol, self).__init__()
@@ -175,20 +190,21 @@ class DillableProtocol(BaseProtocol):
         if self._dill is not None:
             return self._dill
 
-        dill_already_imported = 'dill' in sys.modules
+        dill_already_imported = "dill" in sys.modules
         self._dill = import_optional_dependency(
-            'dill', purpose='the @dillable protocol')
+            "dill", purpose="the @dillable protocol"
+        )
         if not dill_already_imported and self._suppress_dill_side_effects:
             self._dill.extend(False)
 
         return self._dill
 
     def write(self, value, path):
-        with path.open('wb') as file_:
+        with path.open("wb") as file_:
             self._get_dill_module().dump(value, file_)
 
     def read(self, path, extension):
-        with path.open('rb') as file_:
+        with path.open("rb") as file_:
             return self._get_dill_module().load(file_)
 
 
@@ -210,42 +226,48 @@ class ParquetDataFrameProtocol(BaseProtocol):
         self._check_dtypes = check_dtypes
 
     def get_fixed_file_extension(self):
-        return 'pq'
+        return "pq"
 
     def validate(self, value):
         assert isinstance(value, pd.DataFrame)
 
     def read(self, path, extension):
-        with path.open('rb') as file_:
+        with path.open("rb") as file_:
             return parquet.read_table(file_).to_pandas()
 
     def write(self, df, path):
         self._check_no_duplicate_cols(df)
         if self._check_dtypes:
             self._check_no_categorical_cols(df)
-        with path.open('wb') as file_:
+        with path.open("wb") as file_:
             parquet.write_table(Table.from_pandas(df), file_)
 
     def _check_no_duplicate_cols(self, df):
-        duplicate_cols = {elem: count for elem, count in Counter(df.columns).items() if count > 1}
+        duplicate_cols = {
+            elem: count for elem, count in Counter(df.columns).items() if count > 1
+        }
         if duplicate_cols:
             raise ValueError(
-                oneline(f'''
+                oneline(
+                    f"""
                     Attempted to serialize to Parquet a dataframe which has
                     duplicate columns with the following counts:
                     {duplicate_cols}. You can fix this by dropping duplicate
-                    columns with something like: ''')
-                + '\n' +
-                "df = df.loc[:, ~df.columns.duplicated()]")
+                    columns with something like: """
+                )
+                + "\n"
+                + "df = df.loc[:, ~df.columns.duplicated()]"
+            )
 
     def _check_no_categorical_cols(self, df):
         categorical_cols = [
-            col for col in df.columns
-            if df[col].dtype.name == 'category'
+            col for col in df.columns if df[col].dtype.name == "category"
         ]
 
         if categorical_cols:
-            raise ValueError(oneline(f'''
+            raise ValueError(
+                oneline(
+                    f"""
                 Attempted to serialize to Parquet a dataframe which has
                 categorical columns: {categorical_cols!r} --
                 these columns may be transformed to another type and/or
@@ -253,7 +275,9 @@ class ParquetDataFrameProtocol(BaseProtocol):
                 You can fix this by using
                 (a) ``@frame(file_format='feather')`` to use the Feather
                 format instead, or
-                (b) ``@frame(check_dtypes=False)`` to ignore this check.'''))
+                (b) ``@frame(check_dtypes=False)`` to ignore this check."""
+                )
+            )
 
 
 class FeatherDataFrameProtocol(BaseProtocol):
@@ -265,21 +289,21 @@ class FeatherDataFrameProtocol(BaseProtocol):
     """
 
     def get_fixed_file_extension(self):
-        return 'feather'
+        return "feather"
 
     def validate(self, value):
         assert isinstance(value, pd.DataFrame)
 
     def read(self, path, extension):
-        with path.open('rb') as file_:
+        with path.open("rb") as file_:
             return pd.read_feather(file_)
 
     def write(self, df, path):
-        with path.open('wb') as file_:
+        with path.open("wb") as file_:
             df.to_feather(file_)
 
 
-Image = import_optional_dependency('PIL.Image', raise_on_missing=False)
+Image = import_optional_dependency("PIL.Image", raise_on_missing=False)
 
 
 class ImageProtocol(BaseProtocol):
@@ -291,7 +315,7 @@ class ImageProtocol(BaseProtocol):
     """
 
     def get_fixed_file_extension(self):
-        return 'png'
+        return "png"
 
     def validate(self, value):
         # If Image is None, then the PIL library is not present, which
@@ -301,9 +325,10 @@ class ImageProtocol(BaseProtocol):
         assert isinstance(value, Image.Image)
 
     def read(self, path, extension):
-        with path.open('rb') as file_:
+        with path.open("rb") as file_:
             Image = import_optional_dependency(
-                'PIL.Image', purpose='the @image decorator')
+                "PIL.Image", purpose="the @image decorator"
+            )
             image = Image.open(file_)
             # Image.open() is lazy; if we don't call load() now, the file can
             # be closed or possibly invalidated before it actually gets read.
@@ -311,8 +336,8 @@ class ImageProtocol(BaseProtocol):
             return image
 
     def write(self, image, path):
-        with path.open('wb') as file_:
-            image.save(file_, format='png')
+        with path.open("wb") as file_:
+            image.save(file_, format="png")
 
 
 class NumPyProtocol(BaseProtocol):
@@ -324,21 +349,21 @@ class NumPyProtocol(BaseProtocol):
     """
 
     def get_fixed_file_extension(self):
-        return '.npy'
+        return ".npy"
 
     def validate(self, value):
         assert isinstance(value, np.ndarray)
 
     def read(self, path, extension):
-        with path.open('rb') as file_:
+        with path.open("rb") as file_:
             return np.load(file_)
 
     def write(self, array, path):
-        with path.open('wb') as file_:
+        with path.open("wb") as file_:
             np.save(file_, array)
 
 
-dd = import_optional_dependency('dask.dataframe', raise_on_missing=False)
+dd = import_optional_dependency("dask.dataframe", raise_on_missing=False)
 
 
 class DaskProtocol(BaseProtocol):
@@ -346,11 +371,11 @@ class DaskProtocol(BaseProtocol):
     Decorator indicating that an entity's values always have the
     ``dask.dataframe.DataFrame`` type.
 
-    These values will be serialized a .dask.pq directory.
+    These values will be serialized to a .dask.pq directory.
     """
 
     def get_fixed_file_extension(self):
-        return 'pq.dask'
+        return "pq.dask"
 
     def validate(self, value):
         # If dd is None, then dask with dataframe (i.e. dask[dataframe]) is not present,
@@ -360,14 +385,15 @@ class DaskProtocol(BaseProtocol):
         assert isinstance(value, dd.DataFrame)
 
     def read(self, path, extension):
-        dd = import_optional_dependency('dask.dataframe', purpose='the @dask decorator')
+        dd = import_optional_dependency("dask.dataframe", purpose="the @dask decorator")
         with warnings.catch_warnings():
-            warnings.filterwarnings('error', message=r".*cannot\s+autodetect\s+index.*")
+            warnings.filterwarnings("error", message=r".*cannot\s+autodetect\s+index.*")
             try:
                 return dd.read_parquet(path)
             except RuntimeWarning as e:
                 raise UnsupportedSerializedValueError(
-                    f"Reading dataframe failed due to present MultiIndex: {e}")
+                    f"Reading dataframe failed due to present MultiIndex: {e}"
+                )
 
     def write(self, df, path):
         dd.to_parquet(df, path, write_index=True)
@@ -383,19 +409,20 @@ class YamlProtocol(BaseProtocol):
     **kwargs: keyword args for ``yaml.dump``
         E.g. ``default_flow_style``, ``encoding``, etc.
     """
+
     def __init__(self, **kwargs):
         super(YamlProtocol, self).__init__()
         self._kwargs = kwargs
 
     def get_fixed_file_extension(self):
-        return 'yaml'
+        return "yaml"
 
     def write(self, value, path):
-        with path.open('w') as file_:
+        with path.open("w") as file_:
             yaml.dump(value, file_, **self._kwargs)
 
     def read(self, path, extension):
-        with path.open('r') as file_:
+        with path.open("r") as file_:
             return yaml.safe_load(file_)
 
 
@@ -417,17 +444,21 @@ class PathProtocol(BaseProtocol):
         then "copy" is better.
     """
 
-    def __init__(self, operation='copy'):
+    def __init__(self, operation="copy"):
         super(PathProtocol, self).__init__()
-        known_operations = ('move', 'copy')
+        known_operations = ("move", "copy")
         if operation not in known_operations:
-            raise ValueError(oneline(f'''
+            raise ValueError(
+                oneline(
+                    f"""
                 Operation must be in {known_operations!r}:
-                got {operation}.'''))
+                got {operation}."""
+                )
+            )
         self.operation = operation
 
     def get_fixed_file_extension(self):
-        return 'as_path'
+        return "as_path"
 
     def validate(self, value):
         assert isinstance(value, Path)
@@ -444,12 +475,12 @@ class PathProtocol(BaseProtocol):
         dst_dir_path.mkdir()
         dst_path = dst_dir_path / src_path.name
 
-        if self.operation == 'move':
+        if self.operation == "move":
             shutil.move(src_path, dst_path)
-        elif self.operation == 'copy':
+        elif self.operation == "copy":
             recursively_copy_path(src_path, dst_path)
         else:
-            raise AssertionError(f'Unexpected operation: {self.operation!r}')
+            raise AssertionError(f"Unexpected operation: {self.operation!r}")
 
     def read(self, path, extension):
         return single_element(path.iterdir())
@@ -502,7 +533,7 @@ class CombinedProtocol(BaseProtocol):
             return
 
         if not self._subprotocols:
-            raise AssertionError('No subprotocols defined')
+            raise AssertionError("No subprotocols defined")
         else:
             self._subprotocols[-1].validate(value)
 
@@ -510,14 +541,15 @@ class CombinedProtocol(BaseProtocol):
         if not self.can_read_file_extension(extension):
             raise ValueError(
                 "This protocol doesn't know how to read a file with "
-                f"extension {extension!r}")
+                f"extension {extension!r}"
+            )
         return self._protocol_for_extension(extension).read(path, extension)
 
     def write(self, value, path):
         self._protocol_for_value(value).write(value, path)
 
     def __repr__(self):
-        return f'CombinedProtocol{tuple(self._subprotocols)!r}'
+        return f"CombinedProtocol{tuple(self._subprotocols)!r}"
 
 
 class TypeProtocol(PicklableProtocol):
@@ -544,7 +576,7 @@ class TypeProtocol(PicklableProtocol):
         assert isinstance(value, self._type)
 
     def __repr__(self):
-        return f'TypeProtocol({self._type.__name__})'
+        return f"TypeProtocol({self._type.__name__})"
 
 
 class EnumProtocol(PicklableProtocol):
@@ -571,4 +603,4 @@ class EnumProtocol(PicklableProtocol):
         assert value in self._allowed_values
 
     def __repr__(self):
-        return f'EnumProtocol{tuple(self._allowed_values)!r}'
+        return f"EnumProtocol{tuple(self._allowed_values)!r}"
