@@ -3,9 +3,9 @@ import pytest
 import bionic as bn
 
 
-@pytest.fixture(scope='function')
+@pytest.fixture(scope="function")
 def preset_builder(builder):
-    builder.assign('n', values=[1, 2, 3])
+    builder.assign("n", values=[1, 2, 3])
 
     @builder
     def xs(n):
@@ -22,13 +22,13 @@ def test_pyplot_then_gather(preset_builder):
     builder = preset_builder
 
     @builder
-    @bn.pyplot('plt')
-    @bn.gather('n', ['xs', 'ys'])
+    @bn.pyplot("plt")
+    @bn.gather("n", ["xs", "ys"])
     def plot(gather_df, plt):
         for row in gather_df.itertuples():
             plt.plot(row.xs, row.ys)
 
-    img = builder.build().get('plot')
+    img = builder.build().get("plot")
     assert img.width > 0
     assert img.height > 0
 
@@ -37,12 +37,12 @@ def test_gather_then_pyplot(preset_builder):
     builder = preset_builder
 
     @builder
-    @bn.gather('n', ['xs', 'ys'])
-    @bn.pyplot('plt')
+    @bn.gather("n", ["xs", "ys"])
+    @bn.pyplot("plt")
     def plot(gather_df, plt):
         for row in gather_df.itertuples():
             plt.plot(row.xs, row.ys)
 
-    img = builder.build().get('plot')
+    img = builder.build().get("plot")
     assert img.width > 0
     assert img.height > 0
