@@ -3,7 +3,7 @@ import pytest
 import math
 
 from ..helpers import count_calls, ResettingCounter, RoundingProtocol
-from bionic.exception import CodeVersioningError
+from bionic.exception import AttributeValidationError, CodeVersioningError
 
 import bionic as bn
 
@@ -24,9 +24,9 @@ class ReadCountingProtocol(bn.protocols.PicklableProtocol):
 def test_caching_and_invalidation(builder):
     # Set up the builder with singleton values.
 
-    builder.assign('x', 2)
-    builder.assign('y', 3)
-    builder.assign('z', 4)
+    builder.assign("x", 2)
+    builder.assign("y", 3)
+    builder.assign("z", 4)
 
     @builder
     @count_calls
@@ -47,76 +47,80 @@ def test_caching_and_invalidation(builder):
     # Access the downstream values.
     flow = builder.build()
 
-    assert flow.get('xy') == 6
-    assert flow.get('xy') == 6
+    assert flow.get("xy") == 6
+    assert flow.get("xy") == 6
     assert xy.times_called() == 1
 
-    assert flow.get('yz') == 12
-    assert flow.get('yz') == 12
+    assert flow.get("yz") == 12
+    assert flow.get("yz") == 12
     assert yz.times_called() == 1
 
-    assert flow.get('xy_plus_yz') == 18
-    assert flow.get('xy_plus_yz') == 18
+    assert flow.get("xy_plus_yz") == 18
+    assert flow.get("xy_plus_yz") == 18
     assert xy_plus_yz.times_called() == 1
 
     # Rebuild the flow (resetting the in-memory cache) and confirm that
     # xy and xy_plus_yz are still cached.
     flow = builder.build()
 
-    assert flow.get('xy') == 6
+    assert flow.get("xy") == 6
     assert xy.times_called() == 0
 
-    assert flow.get('yz') == 12
+    assert flow.get("yz") == 12
     # Note yz is not cached.
     assert yz.times_called() == 1
 
-    assert flow.get('xy_plus_yz') == 18
+    assert flow.get("xy_plus_yz") == 18
     assert xy_plus_yz.times_called() == 0
 
     # Change the value of z, and confirm that yz and xy_plus_yz are recomputed.
-    flow = flow.setting('z', -4)
+    flow = flow.setting("z", -4)
 
-    assert flow.get('xy') == 6
+    assert flow.get("xy") == 6
     assert xy.times_called() == 0
 
-    assert flow.get('yz') == -12
+    assert flow.get("yz") == -12
     assert yz.times_called() == 1
 
-    assert flow.get('xy_plus_yz') == -6
-    assert flow.get('xy_plus_yz') == -6
+    assert flow.get("xy_plus_yz") == -6
+    assert flow.get("xy_plus_yz") == -6
     assert xy_plus_yz.times_called() == 1
 
     # Update x and y to have multiple values, and confirm that xy and
     # xy_plus_yz are recomputed.
-    flow = builder.build()\
-        .setting('x', values=[2, -2])\
-        .setting('y', values=[3, 6])
+    flow = builder.build().setting("x", values=[2, -2]).setting("y", values=[3, 6])
 
-    assert flow.get('xy', set) == {-2*6, -2*3, 2*3, 2*6}  # noqa: E226
+    assert flow.get("xy", set) == {-2 * 6, -2 * 3, 2 * 3, 2 * 6}  # noqa: E226
     # Note that we only call xy 3 times, because one value was already cached.
     assert xy.times_called() == 3
 
-    assert flow.get('yz', set) == {3*4, 6*4}  # noqa: E226
+    assert flow.get("yz", set) == {3 * 4, 6 * 4}  # noqa: E226
     assert yz.times_called() == 2
 
-    assert flow.get('xy_plus_yz', set) == {
-        -2*3+3*4, -2*6+6*4, 2*3+3*4, 2*6+6*4}  # noqa: E226
+    assert flow.get("xy_plus_yz", set) == {
+        -2 * 3 + 3 * 4,
+        -2 * 6 + 6 * 4,
+        2 * 3 + 3 * 4,
+        2 * 6 + 6 * 4,
+    }  # noqa: E226
     assert xy.times_called() == 0
     assert yz.times_called() == 0
     assert xy_plus_yz.times_called() == 3
 
-    flow = builder.build()\
-        .setting('x', values=[2, -2])\
-        .setting('y', values=[3, 6])
+    flow = builder.build().setting("x", values=[2, -2]).setting("y", values=[3, 6])
 
-    assert flow.get('xy', set) == {-12, -6, 6, 12}
+    assert flow.get("xy", set) == {-12, -6, 6, 12}
     assert xy.times_called() == 0
 
-    assert flow.get('yz', set) == {3*4, 6*4}  # noqa: E226
+    assert flow.get("yz", set) == {3 * 4, 6 * 4}  # noqa: E226
     assert yz.times_called() == 2
 
-    assert flow.get('xy_plus_yz', set) == {
-        -2*3+3*4, -2*6+6*4, 2*3+3*4, 2*6+6*4}  # noqa: E226
+    assert flow.get("xy_plus_yz", set) == {
+        -2 * 3 + 3 * 4,
+        -2 * 6 + 6 * 4,
+        2 * 3 + 3 * 4,
+        2 * 6 + 6 * 4,
+    }  # noqa: E226
     assert xy.times_called() == 0
     assert yz.times_called() == 0
     assert xy_plus_yz.times_called() == 0
@@ -124,51 +128,55 @@ def test_caching_and_invalidation(builder):
     # Update y to have a different, overlapped set of values, and check that
     # the minimal set of recomputations are performed.
 
-    flow = flow.setting('y', values=[6, 9])
+    flow = flow.setting("y", values=[6, 9])
 
-    assert flow.get('xy', set) == {-2*6, -2*9, 2*6, 2*9}  # noqa: E226
+    assert flow.get("xy", set) == {-2 * 6, -2 * 9, 2 * 6, 2 * 9}  # noqa: E226
     assert xy.times_called() == 2
 
-    assert flow.get('yz', set) == {6*4, 9*4}  # noqa: E226
+    assert flow.get("yz", set) == {6 * 4, 9 * 4}  # noqa: E226
     assert yz.times_called() == 2
 
-    assert flow.get('xy_plus_yz', set) == {
-        -2*6+6*4, -2*9+9*4, 2*6+6*4, 2*9+9*4}  # noqa: E226
+    assert flow.get("xy_plus_yz", set) == {
+        -2 * 6 + 6 * 4,
+        -2 * 9 + 9 * 4,
+        2 * 6 + 6 * 4,
+        2 * 9 + 9 * 4,
+    }  # noqa: E226
     assert xy_plus_yz.times_called() == 2
 
     # This is mainly just to check that the cache wrapper returns a sane set of
     # case keys.
-    key_names = flow.get('xy_plus_yz', 'series').index.names
-    for name in ['x', 'y']:
+    key_names = flow.get("xy_plus_yz", "series").index.names
+    for name in ["x", "y"]:
         assert name in key_names
 
 
 def test_versioning(builder):
     call_counter = ResettingCounter()
 
-    builder.assign('x', 2)
-    builder.assign('y', 3)
+    builder.assign("x", 2)
+    builder.assign("y", 3)
 
     @builder
     def f(x, y):
         call_counter.mark()
         return x + y
 
-    assert builder.build().get('f') == 5
-    assert builder.build().get('f') == 5
+    assert builder.build().get("f") == 5
+    assert builder.build().get("f") == 5
     assert call_counter.times_called() == 1
 
-    builder.delete('f')
+    builder.delete("f")
 
     @builder  # noqa: F811
     def f(x, y):
         call_counter.mark()
         return x * y
 
-    assert builder.build().get('f') == 5
+    assert builder.build().get("f") == 5
     assert call_counter.times_called() == 0
 
-    builder.delete('f')
+    builder.delete("f")
 
     @builder  # noqa: F811
     @bn.version(1)
@@ -176,10 +184,10 @@ def test_versioning(builder):
         call_counter.mark()
         return x * y
 
-    assert builder.build().get('f') == 6
+    assert builder.build().get("f") == 6
     assert call_counter.times_called() == 1
 
-    builder.delete('f')
+    builder.delete("f")
 
     @builder  # noqa: F811
     @bn.version(1)
@@ -187,10 +195,10 @@ def test_versioning(builder):
         call_counter.mark()
         return y * x
 
-    assert builder.build().get('f') == 6
+    assert builder.build().get("f") == 6
     assert call_counter.times_called() == 0
 
-    builder.delete('f')
+    builder.delete("f")
 
     @builder  # noqa: F811
     @bn.version(major=1, minor=1)
@@ -198,7 +206,7 @@ def test_versioning(builder):
         call_counter.mark()
         return y * x
 
-    assert builder.build().get('f') == 6
+    assert builder.build().get("f") == 6
     assert call_counter.times_called() == 0
 
     @builder  # noqa: F811
@@ -207,10 +215,10 @@ def test_versioning(builder):
         call_counter.mark()
         return x ** y
 
-    assert builder.build().get('f') == 6
+    assert builder.build().get("f") == 6
     assert call_counter.times_called() == 0
 
-    builder.delete('f')
+    builder.delete("f")
 
     @builder  # noqa: F811
     @bn.version(major=2)
@@ -218,7 +226,7 @@ def test_versioning(builder):
         call_counter.mark()
         return x ** y
 
-    assert builder.build().get('f') == 8
+    assert builder.build().get("f") == 8
     assert call_counter.times_called() == 1
 
 
@@ -226,7 +234,7 @@ def test_indirect_versioning(builder):
     y_call_counter = ResettingCounter()
     f_call_counter = ResettingCounter()
 
-    builder.assign('x', 2)
+    builder.assign("x", 2)
 
     @builder
     def y():
@@ -238,7 +246,7 @@ def test_indirect_versioning(builder):
         f_call_counter.mark()
         return x + y
 
-    assert builder.build().get('f') == 5
+    assert builder.build().get("f") == 5
     assert y_call_counter.times_called() == 1
     assert f_call_counter.times_called() == 1
 
@@ -247,7 +255,7 @@ def test_indirect_versioning(builder):
         y_call_counter.mark()
         return 4
 
-    assert builder.build().get('f') == 5
+    assert builder.build().get("f") == 5
     assert y_call_counter.times_called() == 0
     assert f_call_counter.times_called() == 0
 
@@ -257,7 +265,7 @@ def test_indirect_versioning(builder):
         y_call_counter.mark()
         return 4
 
-    assert builder.build().get('f') == 6
+    assert builder.build().get("f") == 6
     assert y_call_counter.times_called() == 1
     assert f_call_counter.times_called() == 1
 
@@ -265,9 +273,9 @@ def test_indirect_versioning(builder):
     @bn.version(1)
     def y():
         y_call_counter.mark()
-        return len('xxxx')
+        return len("xxxx")
 
-    assert builder.build().get('f') == 6
+    assert builder.build().get("f") == 6
     assert y_call_counter.times_called() == 0
     assert f_call_counter.times_called() == 0
 
@@ -275,21 +283,21 @@ def test_indirect_versioning(builder):
     @bn.version(1, minor=1)
     def y():
         y_call_counter.mark()
-        return len('xxxx')
+        return len("xxxx")
 
-    assert builder.build().get('f') == 6
+    assert builder.build().get("f") == 6
     assert y_call_counter.times_called() == 0
     assert f_call_counter.times_called() == 0
 
-    builder.set('x', 5)
+    builder.set("x", 5)
 
-    assert builder.build().get('f') == 9
+    assert builder.build().get("f") == 9
     assert y_call_counter.times_called() == 0
     assert f_call_counter.times_called() == 1
 
-    builder.set('x', 2)
+    builder.set("x", 2)
 
-    assert builder.build().get('f') == 6
+    assert builder.build().get("f") == 6
     assert y_call_counter.times_called() == 0
     assert f_call_counter.times_called() == 0
 
@@ -297,21 +305,21 @@ def test_indirect_versioning(builder):
 def test_versioning_assist(builder):
     call_counter = ResettingCounter()
 
-    builder.set('core__versioning_mode', 'assist')
+    builder.set("core__versioning_mode", "assist")
 
-    builder.assign('x', 2)
-    builder.assign('y', 3)
+    builder.assign("x", 2)
+    builder.assign("y", 3)
 
     @builder
     def f(x, y):
         call_counter.mark()
         return x + y
 
-    assert builder.build().get('f') == 5
-    assert builder.build().get('f') == 5
+    assert builder.build().get("f") == 5
+    assert builder.build().get("f") == 5
     assert call_counter.times_called() == 1
 
-    builder.delete('f')
+    builder.delete("f")
 
     @builder  # noqa: F811
     def f(x, y):
@@ -319,9 +327,9 @@ def test_versioning_assist(builder):
         return x * y
 
     with pytest.raises(CodeVersioningError):
-        builder.build().get('f')
+        builder.build().get("f")
 
-    builder.delete('f')
+    builder.delete("f")
 
     @builder  # noqa: F811
     @bn.version(1)
@@ -329,10 +337,10 @@ def test_versioning_assist(builder):
         call_counter.mark()
         return x * y
 
-    assert builder.build().get('f') == 6
+    assert builder.build().get("f") == 6
     assert call_counter.times_called() == 1
 
-    builder.delete('f')
+    builder.delete("f")
 
     @builder  # noqa: F811
     @bn.version(1)
@@ -341,9 +349,9 @@ def test_versioning_assist(builder):
         return y * x
 
     with pytest.raises(CodeVersioningError):
-        builder.build().get('f')
+        builder.build().get("f")
 
-    builder.delete('f')
+    builder.delete("f")
 
     @builder  # noqa: F811
     @bn.version(major=1, minor=1)
@@ -351,7 +359,7 @@ def test_versioning_assist(builder):
         call_counter.mark()
         return y * x
 
-    assert builder.build().get('f') == 6
+    assert builder.build().get("f") == 6
     assert call_counter.times_called() == 0
 
     @builder  # noqa: F811
@@ -361,9 +369,9 @@ def test_versioning_assist(builder):
         return x ** y
 
     with pytest.raises(CodeVersioningError):
-        builder.build().get('f')
+        builder.build().get("f")
 
-    builder.delete('f')
+    builder.delete("f")
 
     @builder  # noqa: F811
     @bn.version(major=2)
@@ -371,7 +379,7 @@ def test_versioning_assist(builder):
         call_counter.mark()
         return x ** y
 
-    assert builder.build().get('f') == 8
+    assert builder.build().get("f") == 8
     assert call_counter.times_called() == 1
 
 
@@ -379,9 +387,9 @@ def test_indirect_versioning_assist(builder):
     y_call_counter = ResettingCounter()
     f_call_counter = ResettingCounter()
 
-    builder.set('core__versioning_mode', 'assist')
+    builder.set("core__versioning_mode", "assist")
 
-    builder.assign('x', 2)
+    builder.assign("x", 2)
 
     @builder
     def y():
@@ -393,7 +401,7 @@ def test_indirect_versioning_assist(builder):
         f_call_counter.mark()
         return x + y
 
-    assert builder.build().get('f') == 5
+    assert builder.build().get("f") == 5
     assert y_call_counter.times_called() == 1
     assert f_call_counter.times_called() == 1
 
@@ -403,7 +411,7 @@ def test_indirect_versioning_assist(builder):
         return 4
 
     with pytest.raises(CodeVersioningError):
-        builder.build().get('f')
+        builder.build().get("f")
 
     @builder  # noqa: F811
     @bn.version(1)
@@ -411,7 +419,7 @@ def test_indirect_versioning_assist(builder):
         y_call_counter.mark()
         return 4
 
-    assert builder.build().get('f') == 6
+    assert builder.build().get("f") == 6
     assert y_call_counter.times_called() == 1
     assert f_call_counter.times_called() == 1
 
@@ -419,30 +427,30 @@ def test_indirect_versioning_assist(builder):
     @bn.version(1)
     def y():
         y_call_counter.mark()
-        return len('xxxx')
+        return len("xxxx")
 
     with pytest.raises(CodeVersioningError):
-        builder.build().get('f')
+        builder.build().get("f")
 
     @builder  # noqa: F811
     @bn.version(1, minor=1)
     def y():
         y_call_counter.mark()
-        return len('xxxx')
+        return len("xxxx")
 
-    assert builder.build().get('f') == 6
+    assert builder.build().get("f") == 6
     assert y_call_counter.times_called() == 0
     assert f_call_counter.times_called() == 0
 
-    builder.set('x', 5)
+    builder.set("x", 5)
 
-    assert builder.build().get('f') == 9
+    assert builder.build().get("f") == 9
     assert y_call_counter.times_called() == 0
     assert f_call_counter.times_called() == 1
 
-    builder.set('x', 2)
+    builder.set("x", 2)
 
-    assert builder.build().get('f') == 6
+    assert builder.build().get("f") == 6
     assert y_call_counter.times_called() == 0
     assert f_call_counter.times_called() == 0
 
@@ -450,31 +458,31 @@ def test_indirect_versioning_assist(builder):
 def test_versioning_auto(builder):
     call_counter = ResettingCounter()
 
-    builder.set('core__versioning_mode', 'auto')
+    builder.set("core__versioning_mode", "auto")
 
-    builder.assign('x', 2)
-    builder.assign('y', 3)
+    builder.assign("x", 2)
+    builder.assign("y", 3)
 
     @builder
     def f(x, y):
         call_counter.mark()
         return x + y
 
-    assert builder.build().get('f') == 5
-    assert builder.build().get('f') == 5
+    assert builder.build().get("f") == 5
+    assert builder.build().get("f") == 5
     assert call_counter.times_called() == 1
 
-    builder.delete('f')
+    builder.delete("f")
 
     @builder  # noqa: F811
     def f(x, y):
         call_counter.mark()
         return x * y
 
-    assert builder.build().get('f') == 6
+    assert builder.build().get("f") == 6
     assert call_counter.times_called() == 1
 
-    builder.delete('f')
+    builder.delete("f")
 
     @builder  # noqa: F811
     @bn.version(1)
@@ -482,10 +490,10 @@ def test_versioning_auto(builder):
         call_counter.mark()
         return x * y
 
-    assert builder.build().get('f') == 6
+    assert builder.build().get("f") == 6
     assert call_counter.times_called() == 1
 
-    builder.delete('f')
+    builder.delete("f")
 
     @builder  # noqa: F811
     @bn.version(1)
@@ -493,10 +501,10 @@ def test_versioning_auto(builder):
         call_counter.mark()
         return y * x
 
-    assert builder.build().get('f') == 6
+    assert builder.build().get("f") == 6
     assert call_counter.times_called() == 1
 
-    builder.delete('f')
+    builder.delete("f")
 
     @builder  # noqa: F811
     @bn.version(major=1, minor=1)
@@ -504,7 +512,7 @@ def test_versioning_auto(builder):
         call_counter.mark()
         return y * x
 
-    assert builder.build().get('f') == 6
+    assert builder.build().get("f") == 6
     assert call_counter.times_called() == 0
 
     @builder  # noqa: F811
@@ -513,10 +521,10 @@ def test_versioning_auto(builder):
         call_counter.mark()
         return x ** y
 
-    assert builder.build().get('f') == 8
+    assert builder.build().get("f") == 8
     assert call_counter.times_called() == 1
 
-    builder.delete('f')
+    builder.delete("f")
 
     @builder  # noqa: F811
     @bn.version(major=2)
@@ -524,7 +532,7 @@ def test_versioning_auto(builder):
         call_counter.mark()
         return x ** y
 
-    assert builder.build().get('f') == 8
+    assert builder.build().get("f") == 8
     assert call_counter.times_called() == 1
 
 
@@ -532,9 +540,9 @@ def test_indirect_versioning_auto(builder):
     y_call_counter = ResettingCounter()
     f_call_counter = ResettingCounter()
 
-    builder.set('core__versioning_mode', 'auto')
+    builder.set("core__versioning_mode", "auto")
 
-    builder.assign('x', 2)
+    builder.assign("x", 2)
 
     @builder
     def y():
@@ -546,7 +554,7 @@ def test_indirect_versioning_auto(builder):
         f_call_counter.mark()
         return x + y
 
-    assert builder.build().get('f') == 5
+    assert builder.build().get("f") == 5
     assert y_call_counter.times_called() == 1
     assert f_call_counter.times_called() == 1
 
@@ -555,7 +563,7 @@ def test_indirect_versioning_auto(builder):
         y_call_counter.mark()
         return 4
 
-    assert builder.build().get('f') == 6
+    assert builder.build().get("f") == 6
     assert y_call_counter.times_called() == 1
     assert f_call_counter.times_called() == 1
 
@@ -565,39 +573,43 @@ def test_indirect_versioning_auto(builder):
         y_call_counter.mark()
         return 4
 
-    assert builder.build().get('f') == 6
+    assert builder.build().get("f") == 6
     assert y_call_counter.times_called() == 1
-    assert f_call_counter.times_called() == 1
+    # f uses the cached value since final values
+    # of x and y are still the same
+    assert f_call_counter.times_called() == 0
 
     @builder  # noqa: F811
     @bn.version(1)
     def y():
         y_call_counter.mark()
-        return len('xxxx')
+        return len("xxxx")
 
-    assert builder.build().get('f') == 6
+    assert builder.build().get("f") == 6
     assert y_call_counter.times_called() == 1
-    assert f_call_counter.times_called() == 1
+    # f uses the cached value since final values
+    # of x and y are still the same
+    assert f_call_counter.times_called() == 0
 
     @builder  # noqa: F811
     @bn.version(1, minor=1)
     def y():
         y_call_counter.mark()
-        return len('xxxx')
+        return len("xxxx")
 
-    assert builder.build().get('f') == 6
+    assert builder.build().get("f") == 6
     assert y_call_counter.times_called() == 0
     assert f_call_counter.times_called() == 0
 
-    builder.set('x', 5)
+    builder.set("x", 5)
 
-    assert builder.build().get('f') == 9
+    assert builder.build().get("f") == 9
     assert y_call_counter.times_called() == 0
     assert f_call_counter.times_called() == 1
 
-    builder.set('x', 2)
+    builder.set("x", 2)
 
-    assert builder.build().get('f') == 6
+    assert builder.build().get("f") == 6
     assert y_call_counter.times_called() == 0
     assert f_call_counter.times_called() == 0
 
@@ -609,9 +621,9 @@ def test_all_returned_results_are_deserialized(builder):
     def pi():
         return math.pi
 
-    assert builder.build().get('pi') == 3
-    assert builder.build().get('pi') == 3
-    assert builder.build().get('pi') != math.pi
+    assert builder.build().get("pi") == 3
+    assert builder.build().get("pi") == 3
+    assert builder.build().get("pi") != math.pi
     assert pi.times_called() == 1
 
 
@@ -619,7 +631,7 @@ def test_deps_of_cached_values_not_needed(builder):
     y_protocol = ReadCountingProtocol()
     z_protocol = ReadCountingProtocol()
 
-    builder.assign('x', 2)
+    builder.assign("x", 2)
 
     @builder
     @y_protocol
@@ -632,71 +644,111 @@ def test_deps_of_cached_values_not_needed(builder):
         return y + 1
 
     flow = builder.build()
-    assert flow.get('x') == 2
-    assert flow.get('y') == 3
-    assert flow.get('z') == 4
+    assert flow.get("x") == 2
+    assert flow.get("y") == 3
+    assert flow.get("z") == 4
 
-    assert flow.get('x') == 2
-    assert flow.get('y') == 3
-    assert flow.get('z') == 4
+    assert flow.get("x") == 2
+    assert flow.get("y") == 3
+    assert flow.get("z") == 4
 
     assert y_protocol.times_read_called == 1
     assert z_protocol.times_read_called == 1
 
     flow = builder.build()
-    assert flow.get('z') == 4
+    assert flow.get("z") == 4
 
     assert y_protocol.times_read_called == 1
     assert z_protocol.times_read_called == 2
 
 
-def test_gather_cache_invalidation(builder):
-    builder.assign('x', values=[1, 2])
-    builder.assign('y', values=[2, 3])
+def test_deps_not_called_when_values_not_changed(builder):
+    builder.assign("x", 2)
+    builder.assign("y", 3)
+    builder.assign("z", 4)
 
     @builder
-    @bn.gather('x', 'x', 'df')
+    @count_calls
+    def xy(x, y):
+        return x * y
+
+    @builder
+    @count_calls
+    def yz(y, z):
+        return y * z
+
+    @builder
+    @count_calls
+    def xy_plus_yz(xy, yz):
+        return xy + yz
+
+    flow = builder.build()
+    assert flow.get("xy") == 6
+    assert flow.get("yz") == 12
+    assert flow.get("xy_plus_yz") == 18
+
+    assert xy.times_called() == 1
+    assert yz.times_called() == 1
+    assert xy_plus_yz.times_called() == 1
+
+    flow = flow.setting("x", 1).setting("y", 6).setting("z", 2)
+    assert flow.get("xy") == 6
+    assert flow.get("yz") == 12
+    assert flow.get("xy_plus_yz") == 18
+
+    # xy_plus_yz should not be called again
+    assert xy.times_called() == 1
+    assert yz.times_called() == 1
+    assert xy_plus_yz.times_called() == 0
+
+
+def test_gather_cache_invalidation(builder):
+    builder.assign("x", values=[1, 2])
+    builder.assign("y", values=[2, 3])
+
+    @builder
+    @bn.gather("x", "x", "df")
     @count_calls
     def z(df, y):
-        return df['x'].sum() + y
+        return df["x"].sum() + y
 
-    assert builder.build().get('z', set) == {5, 6}
+    assert builder.build().get("z", set) == {5, 6}
     assert z.times_called() == 2
-    assert builder.build().get('z', set) == {5, 6}
+    assert builder.build().get("z", set) == {5, 6}
     assert z.times_called() == 0
 
-    assert builder.build().setting('x', values=[2, 3]).get('z', set) == {7, 8}
+    assert builder.build().setting("x", values=[2, 3]).get("z", set) == {7, 8}
     assert z.times_called() == 2
 
-    builder.set('y', values=[3, 4])
+    builder.set("y", values=[3, 4])
 
-    assert builder.build().get('z', set) == {6, 7}
+    assert builder.build().get("z", set) == {6, 7}
     assert z.times_called() == 1
 
 
 def test_gather_cache_invalidation_with_over_vars(builder):
-    builder.assign('x', values=[1, 2])
-    builder.assign('y', values=[2, 3])
+    builder.assign("x", values=[1, 2])
+    builder.assign("y", values=[2, 3])
 
     @builder
-    @bn.gather('x', 'y', 'df')
+    @bn.gather("x", "y", "df")
     @count_calls
     def z(df):
         return df.sum().sum()
 
-    assert builder.build().get('z', set) == {7, 9}
+    assert builder.build().get("z", set) == {7, 9}
     assert z.times_called() == 2
-    assert builder.build().get('z', set) == {7, 9}
+    assert builder.build().get("z", set) == {7, 9}
     assert z.times_called() == 0
 
     # If we change one of the values of `x`, both values of `z` should change
     # (because each instance depends on both values of `x`).
-    assert builder.build().setting('x', values=[2, 3]).get('z', set) == {9, 11}
+    assert builder.build().setting("x", values=[2, 3]).get("z", set) == {9, 11}
     assert z.times_called() == 2
 
     # If we change one of the values of `y`, only one value of `z` should
     # change.
-    assert builder.build().setting('y', values=[3, 4]).get('z', set) == {9, 11}
+    assert builder.build().setting("y", values=[3, 4]).get("z", set) == {9, 11}
     assert z.times_called() == 1
 
 
@@ -707,7 +759,7 @@ class Point(object):
 
 
 def test_complex_input_type(builder):
-    builder.assign('point', Point(2, 3))
+    builder.assign("point", Point(2, 3))
 
     @builder
     def x(point):
@@ -724,18 +776,18 @@ def test_complex_input_type(builder):
 
     flow = builder.build()
 
-    assert flow.get('x_plus_y') == 5
+    assert flow.get("x_plus_y") == 5
     assert x_plus_y.times_called() == 1
-    assert flow.get('x_plus_y') == 5
+    assert flow.get("x_plus_y") == 5
     assert x_plus_y.times_called() == 0
 
     builder = flow.to_builder()
-    builder.set('point', values=(Point(2, 3), Point(4, 5)))
+    builder.set("point", values=(Point(2, 3), Point(4, 5)))
     flow = builder.build()
 
-    assert flow.get('x_plus_y', set) == {5, 9}
+    assert flow.get("x_plus_y", set) == {5, 9}
     assert x_plus_y.times_called() == 1
-    assert flow.get('x_plus_y', set) == {5, 9}
+    assert flow.get("x_plus_y", set) == {5, 9}
     assert x_plus_y.times_called() == 0
 
 
@@ -745,8 +797,8 @@ def test_persisting_none(builder):
     def none():
         return None
 
-    assert builder.build().get('none') is None
-    assert builder.build().get('none') is None
+    assert builder.build().get("none") is None
+    assert builder.build().get("none") is None
     assert none.times_called() == 1
 
 
@@ -760,11 +812,12 @@ def test_disable_memory_caching(builder):
         return 1
 
     flow = builder.build()
-    assert flow.get('x') == 1
-    assert flow.get('x') == 1
+    assert flow.get("x") == 1
+    assert flow.get("x") == 1
     assert x_protocol.times_read_called == 2
 
     with pytest.raises(ValueError):
+
         @builder
         @x_protocol
         @bn.persist(False)
@@ -773,4 +826,117 @@ def test_disable_memory_caching(builder):
             return 1
 
         flow = builder.build()
-        assert flow.get('y') == 1
+        assert flow.get("y") == 1
+
+
+def test_changes_per_run_and_not_persist(builder):
+    builder.assign("x", 5)
+
+    @builder
+    @bn.persist(False)
+    @bn.changes_per_run
+    @count_calls
+    def x_plus_one(x):
+        return x + 1
+
+    @builder
+    @bn.persist(False)
+    @count_calls
+    def x_plus_two(x_plus_one):
+        return x_plus_one + 1
+
+    @builder
+    @count_calls
+    def x_plus_three(x_plus_two):
+        return x_plus_two + 1
+
+    @builder
+    @count_calls
+    def x_plus_four(x_plus_three):
+        return x_plus_three + 1
+
+    flow = builder.build()
+    assert flow.get("x_plus_four") == 9
+    assert x_plus_one.times_called() == 1
+    assert x_plus_two.times_called() == 1
+    assert x_plus_three.times_called() == 1
+    assert x_plus_four.times_called() == 1
+    # In the same flow, a nondeterministic entity is not recomputed.
+    assert flow.get("x_plus_one") == 6
+    assert flow.get("x_plus_four") == 9
+    assert x_plus_one.times_called() == 0
+    assert x_plus_two.times_called() == 0
+    assert x_plus_three.times_called() == 0
+    assert x_plus_four.times_called() == 0
+
+    flow = builder.build()
+    assert flow.get("x_plus_four") == 9
+    # x_plus_one changes per run and should be recomputed between runs.
+    assert x_plus_one.times_called() == 1
+    # x_plus_two is a child of a nondeterministic parent which does not persist
+    # and should also be recomputed between runs.
+    assert x_plus_two.times_called() == 1
+    # x_plus_three is a child of a parent that has a nondeterministic parent
+    # and is not persisted. Hence the provenance also contains the
+    # noise from nondeterministic ancestor, cannot use the persisted value and
+    # should also be recomputed.
+    assert x_plus_three.times_called() == 1
+    # x_plus_four uses the value hash of its parent which didn't change.
+    # Should not be recomputed and should use cached value instead.
+    assert x_plus_four.times_called() == 0
+
+
+def test_changes_per_run_and_persist(builder):
+    builder.assign("x", 5)
+
+    @builder
+    @bn.changes_per_run
+    @count_calls
+    def x_plus_one(x):
+        return x + 1
+
+    @builder
+    @bn.persist(False)
+    @count_calls
+    def x_plus_two(x_plus_one):
+        return x_plus_one + 1
+
+    @builder
+    @count_calls
+    def x_plus_three(x_plus_two):
+        return x_plus_two + 1
+
+    flow = builder.build()
+    assert flow.get("x_plus_three") == 8
+    assert x_plus_one.times_called() == 1
+    assert x_plus_two.times_called() == 1
+    assert x_plus_three.times_called() == 1
+    # In the same flow, a nondeterministic entity is not recomputed.
+    assert flow.get("x_plus_one") == 6
+    assert flow.get("x_plus_three") == 8
+    assert x_plus_one.times_called() == 0
+    assert x_plus_two.times_called() == 0
+    assert x_plus_three.times_called() == 0
+
+    flow = builder.build()
+    assert flow.get("x_plus_three") == 8
+    # x_plus_one changes per run and should be recomputed between runs.
+    assert x_plus_one.times_called() == 1
+    # Since the value does not persist, x_plus_two is recomputed.
+    assert x_plus_two.times_called() == 1
+    # Since value of x_plus_one didn't change even though it changes per run,
+    # x_plus_three is not computed.
+    assert x_plus_three.times_called() == 0
+
+
+def test_changes_per_run_and_not_memoize(builder):
+    builder.assign("x", 5)
+
+    @builder
+    @bn.memoize(False)
+    @bn.changes_per_run
+    def x_plus_one(x):
+        return x + 1
+
+    with pytest.raises(AttributeValidationError):
+        builder.build().get("x_plus_one")
